@@ -1,3 +1,4 @@
+#define _DEFAULT_SOURCE
 /*
  * This is free software; you can redistribute it and/or modify it under
  * the terms of the GNU Library General Public License as published by
@@ -31,7 +32,7 @@
  * file except a parse error.
  */
 #define TILDA_CONFIG_OTHER_ERROR \
-    "<b>An unexpected error occured while parsing the config file.</b>\n\n" \
+    "<b>An unexpected error occurred while parsing the config file.</b>\n\n" \
     "The default configuration will be used instead. This error can occur if the configuration file is corrupted" \
     "or otherwise unreadable. Tilda will now start with a default configuration."
 
@@ -40,13 +41,14 @@
 
 #include "debug.h"
 #include "tilda.h"
-#include "callback_func.h"
 #include "configsys.h"
 #include "tilda_window.h"
 #include "key_grabber.h" /* for pull */
 #include "wizard.h"
 #include "xerror.h"
 #include "tomboykeybinder.h"
+#include "tilda-keybinding.h"
+#include "tilda-cli-options.h"
 
 #include <sys/ioctl.h>
 #include <sys/stat.h>
@@ -268,6 +270,12 @@ static gint remove_stale_lock_files ()
     gchar *filename;
     GDir *dir;
 
+    /* if the lock dir does not exist then there are no stale lock files to remove. */
+    if (!g_file_test (lock_dir, G_FILE_TEST_EXISTS)) {
+        g_free (lock_dir);
+        return 0;
+    }
+
     /* Open the lock directory for reading */
     dir = g_dir_open (lock_dir, 0, NULL);
 
@@ -304,182 +312,68 @@ static gint remove_stale_lock_files ()
 }
 
 /**
- * Parse all of the Command-Line Options given to tilda.
- * This can modify argv and argc, and will set values in the config.
+ * Set values in the config from command-line parameters
  *
- * @param argc argc from main
- * @param argv argv from main
- * @return TRUE if we should show the configuration wizard, FALSE otherwise
+ * @param cli_options pointer to a struct containing command-line options
  */
-static gboolean parse_cli (int argc, char *argv[])
+static void setup_config_from_cli_options(tilda_cli_options *cli_options)
 {
-    DEBUG_FUNCTION ("parse_cli");
-    DEBUG_ASSERT (argc != 0);
-    DEBUG_ASSERT (argv != NULL);
+    if (cli_options->background_color != NULL
+            && cli_options->background_color != config_getstr ("background_color")) {
+        config_setstr ("background_color", cli_options->background_color);
 
-    /* Set default values */
-    gchar *background_color = config_getstr ("background_color");
-    gchar *command = config_getstr ("command");
-    gchar *font = config_getstr ("font");
-    gchar *working_dir = config_getstr ("working_dir");
-
-#ifdef VTE_290
-    gchar *image = config_getstr ("image");
-    gint transparency = config_getint ("transparency");
-#else
-    gint back_alpha = config_getint ("back_alpha");
-#endif
-
-    gint lines = config_getint ("lines");
-    gint x_pos = config_getint ("x_pos");
-    gint y_pos = config_getint ("y_pos");
-
-    gboolean antialias = config_getbool ("antialias");
-    gboolean scrollbar = config_getbool ("scrollbar");
-    gboolean show_config = FALSE;
-    gboolean version = FALSE;
-    gboolean hidden = config_getbool ("hidden");
-
-    /* All of the various command-line options */
-    GOptionEntry cl_opts[] = {
-        { "antialias",          'a', 0, G_OPTION_ARG_NONE,      &antialias,         N_("Use Antialiased Fonts"), NULL },
-        { "background-color",   'b', 0, G_OPTION_ARG_STRING,    &background_color,  N_("Set the background color"), NULL },
-        { "command",            'c', 0, G_OPTION_ARG_STRING,    &command,           N_("Run a command at startup"), NULL },
-        { "hidden",             'h', 0, G_OPTION_ARG_NONE,      &hidden,            N_("Start Tilda hidden"), NULL },
-        { "font",               'f', 0, G_OPTION_ARG_STRING,    &font,              N_("Set the font to the following string"), NULL },
-        { "lines",              'l', 0, G_OPTION_ARG_INT,       &lines,             N_("Scrollback Lines"), NULL },
-        { "scrollbar",          's', 0, G_OPTION_ARG_NONE,      &scrollbar,         N_("Use Scrollbar"), NULL },
-        { "version",            'v', 0, G_OPTION_ARG_NONE,      &version,           N_("Print the version, then exit"), NULL },
-        { "working-dir",        'w', 0, G_OPTION_ARG_STRING,    &working_dir,       N_("Set Initial Working Directory"), NULL },
-        { "x-pos",              'x', 0, G_OPTION_ARG_INT,       &x_pos,             N_("X Position"), NULL },
-        { "y-pos",              'y', 0, G_OPTION_ARG_INT,       &y_pos,             N_("Y Position"), NULL },
-#ifdef VTE_290
-        { "image",              'B', 0, G_OPTION_ARG_STRING,    &image,             N_("Set Background Image"), NULL },
-        { "transparency",       't', 0, G_OPTION_ARG_INT,       &transparency,      N_("Opaqueness: 0-100%"), NULL },
-#else
-        { "background-alpha",   't', 0, G_OPTION_ARG_INT,       &back_alpha,        N_("Opaqueness: 0-100%"), NULL },
-#endif
-        { "config",             'C', 0, G_OPTION_ARG_NONE,      &show_config,       N_("Show Configuration Wizard"), NULL },
-        { NULL }
-    };
-
-
-    /* Set up the command-line parser */
-    GError *error = NULL;
-    GOptionContext *context = g_option_context_new (NULL);
-    g_option_context_add_main_entries (context, cl_opts, NULL);
-    g_option_context_parse (context, &argc, &argv, &error);
-    g_option_context_free (context);
-
-    /* Check for unknown options, and give a nice message if there are some */
-    if (error)
-    {
-        g_printerr (_("Error parsing command-line options. Try \"tilda --help\"\nto see all possible options.\n\nError message: %s\n"),
-                    error->message);
-
-        exit (EXIT_FAILURE);
-    }
-
-    /* If we need to show the version, show it then exit normally */
-    if (version)
-    {
-        g_print ("%s\n\n", TILDA_VERSION);
-
-        g_print ("Copyright (c) 2012-2013 Sebastian Geiger (lanoxx@gmx.net)\n");
-        g_print ("Copyright (c) 2005-2009 Tristan Sloughter (sloutri@iit.edu)\n");
-        g_print ("Copyright (c) 2005-2009 Ira W. Snyder (tilda@irasnyder.com)\n\n");
-
-        g_print ("General Information: https://github.com/lanoxx/tilda\n");
-        g_print ("Bug Reports: https://github.com/lanoxx/tilda/issues?state=open\n\n");
-
-        g_print ("This program comes with ABSOLUTELY NO WARRANTY.\n");
-        g_print ("This is free software, and you are welcome to redistribute it\n");
-        g_print ("under certain conditions. See the file COPYING for details.\n");
-
-        exit (EXIT_SUCCESS);
-    }
-
-    /* This block is only used to initialize the Glib and GTK internal options. That way the users can pass additional command line options,
-     * that are used by Glib and GTK. We do this separate from the above options, because we pass TRUE to the gtk_get_option group function
-     * which causes GTK to initialize the default display. This way it is possible to invoke `tilda --version` without getting an
-     * error if there is no display available.
-     */
-    error = NULL;
-    context = g_option_context_new (NULL);
-    g_option_context_add_group (context, gtk_get_option_group (TRUE));
-    g_option_context_parse (context, &argc, &argv, &error);
-    g_option_context_free (context);
-
-    if (error)
-    {
-        g_printerr (_("Error parsing Glib and GTK specific command-line options. Try \"tilda --help-all\"\nto see all possible options.\n\nError message: %s\n"),
-                    error->message);
-
-        exit (EXIT_FAILURE);
-    }
-
-    /* Now set the options in the config, if they changed */
-    if (background_color != config_getstr ("background_color")) {
-        config_setstr ("background_color", background_color);
-
-        GdkColor col;
-        if (gdk_color_parse(background_color, &col)) {
-            config_setint("back_red", col.red);
-            config_setint("back_green", col.green);
-            config_setint("back_blue", col.blue);
+        GdkRGBA col;
+        if (gdk_rgba_parse (&col, cli_options->background_color))
+        {
+            config_setint("back_red", GUINT16_FROM_FLOAT (col.red));
+            config_setint("back_green", GUINT16_FROM_FLOAT (col.green));
+            config_setint("back_blue", GUINT16_FROM_FLOAT (col.blue));
         }
 
-        g_free(background_color);
+        g_free(cli_options->background_color);
     }
-    if (command != config_getstr ("command"))
+    if (cli_options->command != NULL
+            && cli_options->command != config_getstr ("command"))
     {
         config_setbool ("run_command", TRUE);
-        config_setstr ("command", command);
-        g_free(command);
+        config_setstr ("command", cli_options->command);
+        g_free(cli_options->command);
     }
-    if (font != config_getstr ("font")) {
-        config_setstr ("font", font);
-        g_free(font);
+    if (cli_options->font != NULL
+            && cli_options->font != config_getstr ("font")) {
+        config_setstr ("font", cli_options->font);
+        g_free(cli_options->font);
     }
-#ifdef VTE_290
-    if (image != config_getstr ("image")) {
-        config_setstr ("image", image);
-        g_free(image);
-    }
-    if (transparency != config_getint ("transparency"))
+
+    if (cli_options->back_alpha != 0
+            && cli_options->back_alpha != config_getint ("back_alpha"))
     {
-        config_setbool ("enable_transparency", transparency);
-        config_setint ("transparency", transparency);
-    }
-#else
-    if (back_alpha != config_getint ("back_alpha"))
-    {
-        config_setbool ("enable_transparency", ~back_alpha & 0xffff);
-        config_setint ("back_alpha", back_alpha);
-    }
-#endif
-    if (working_dir != config_getstr ("working_dir")) {
-        config_setstr ("working_dir", working_dir);
-        g_free(working_dir);
+        config_setbool ("enable_transparency", ~cli_options->back_alpha & 0xffff);
+        config_setint ("back_alpha", cli_options->back_alpha);
     }
 
-    if (lines != config_getint ("lines"))
-        config_setint ("lines", lines);
+    if (cli_options->working_dir != NULL
+            && cli_options->working_dir != config_getstr ("working_dir")) {
+        config_setstr ("working_dir", cli_options->working_dir);
+        g_free(cli_options->working_dir);
+    }
 
-    if (x_pos != config_getint ("x_pos"))
-        config_setint ("x_pos", x_pos);
-    if (y_pos != config_getint ("y_pos"))
-        config_setint ("y_pos", y_pos);
+    if (cli_options->lines != 0
+            && cli_options->lines != config_getint ("lines"))
+        config_setint ("lines", cli_options->lines);
+    if (cli_options->x_pos != 0
+            && cli_options->x_pos != config_getint ("x_pos"))
+        config_setint ("x_pos", cli_options->x_pos);
+    if (cli_options->y_pos != 0
+            && cli_options->y_pos != config_getint ("y_pos"))
+        config_setint ("y_pos", cli_options->y_pos);
 
-    if (antialias != config_getbool ("antialias"))
-        config_setbool ("antialias", antialias);
-    if (hidden != config_getbool ("hidden"))
-        config_setbool ("hidden", hidden);
-    if (scrollbar != config_getbool ("scrollbar"))
-        config_setbool ("scrollbar", scrollbar);
-
-    /* TRUE if we should show the config wizard, FALSE otherwize */
-    return show_config;
+    if (cli_options->hidden != FALSE
+            && cli_options->hidden != config_getbool ("hidden"))
+        config_setbool ("hidden", cli_options->hidden);
+    if (cli_options->scrollbar != FALSE
+            && cli_options->scrollbar != config_getbool ("scrollbar"))
+        config_setbool ("scrollbar", cli_options->scrollbar);
 }
 
 /**
@@ -509,11 +403,27 @@ static gchar *get_config_file_name (gint instance)
     config_file = g_strdup_printf ("%s%d", config_file_prefix, instance);
     g_free (config_file_prefix);
 
+    /* resolve possible symlink pointing to a different location. */
+    gchar * resolved_config_file = realpath (config_file, NULL);
+
+    if (g_strcmp0 (config_file, resolved_config_file) != 0) {
+        g_debug ("Config file at '%s' points to '%s'", config_file, resolved_config_file);
+    }
+
+    if (resolved_config_file != NULL) {
+        g_free (config_file);
+        /* resolved_config_file must be freed with free(3) so we duplicate the
+         * string to ensure that what ever we return from this function can be
+         * freed with g_free(). */
+        config_file = g_strdup (resolved_config_file);
+        free (resolved_config_file);
+    }
+
     return config_file;
 }
 
-static gint _cmp_locks(gint a, gint b) {
-  return a - b;
+static gint _cmp_locks(gconstpointer a, gconstpointer b, gpointer userdata) {
+  return GPOINTER_TO_INT (a) - GPOINTER_TO_INT (b);
 }
 
 /**
@@ -560,7 +470,7 @@ static gint get_instance_number ()
 
         if (lock != NULL)
         {
-            g_sequence_insert_sorted(seq, GINT_TO_POINTER(lock->instance), (GCompareDataFunc)_cmp_locks, NULL);
+            g_sequence_insert_sorted(seq, GINT_TO_POINTER(lock->instance), _cmp_locks, NULL);
             g_free (lock);
         }
     }
@@ -609,31 +519,80 @@ static void migrate_config_files(char *old_config_path) {
     g_free(new_config_dir);
 }
 
+static void load_application_css () {
+    GtkCssProvider *provider;
+    const gchar* style;
+    GError *error;
+
+    provider = gtk_css_provider_new ();
+
+    gtk_style_context_add_provider_for_screen (gdk_screen_get_default (),
+                                               GTK_STYLE_PROVIDER (provider),
+                                               GTK_STYLE_PROVIDER_PRIORITY_USER);
+
+    style = "#search{background:#fff;}";
+    error = NULL;
+    gtk_css_provider_load_from_data (provider, style, -1, &error);
+
+    if (error) {
+        g_print ("Error %s\n", error->message);
+        g_error_free (error);
+    }
+}
+
 static void load_custom_css_file () {
     GtkCssProvider *provider;
-    char* cssfilename = g_build_filename(
-        g_get_user_config_dir (), "tilda", "style.css", NULL);
-    if (g_file_test (cssfilename, G_FILE_TEST_EXISTS)) {
-        g_print (_("Found style.css in the user config directory, "
-            "applying user css style.\n"));
-        provider = gtk_css_provider_new ();
-        gtk_style_context_add_provider_for_screen (
-            gdk_screen_get_default(),
-            GTK_STYLE_PROVIDER(provider),
-            GTK_STYLE_PROVIDER_PRIORITY_USER
-        );
-        GFile *cssfile = g_file_new_for_path (cssfilename);
-        gtk_css_provider_load_from_file (GTK_CSS_PROVIDER (provider),
-            cssfile, NULL);
-        g_object_unref (cssfile);
+    GFile *file;
+    GError *error;
+    gchar *filename;
+
+    filename = g_build_filename (g_get_user_config_dir (),
+                                 "tilda", "style.css", NULL);
+
+    if (!g_file_test (filename, G_FILE_TEST_EXISTS))
+        return;
+
+    g_print (_("Found style.css in the user config directory, "
+               "applying user css style.\n"));
+
+    provider = gtk_css_provider_new ();
+
+    gtk_style_context_add_provider_for_screen (gdk_screen_get_default(),
+                                               GTK_STYLE_PROVIDER(provider),
+                                               GTK_STYLE_PROVIDER_PRIORITY_USER);
+
+
+    file = g_file_new_for_path (filename);
+    error = NULL;
+    gtk_css_provider_load_from_file (provider, file, &error);
+
+    if (error) {
+        g_print ("Error: %s", error->message);
+        g_error_free (error);
     }
+
+    g_object_unref (file);
+    g_free (filename);
 }
 
 int main (int argc, char *argv[])
 {
-    DEBUG_FUNCTION_MESSAGE ("main", "Using libvte version: %i.%i.%i\n",
+#ifdef DEBUG
+    /**
+     * This enables the tilda log domain while we are in debug mode. This
+     * is for convenience such that we do not need to start tilda with
+     * 'G_MESSAGES_DEBUG=tilda ./tilda' each time we start tilda.
+     */
+
+    g_setenv ("G_MESSAGES_DEBUG", "tilda", FALSE);
+#endif
+
+    DEBUG_FUNCTION_MESSAGE ("main", "Using libvte version: %i.%i.%i",
                             VTE_MAJOR_VERSION, VTE_MINOR_VERSION, VTE_MICRO_VERSION);
 
+    /* Set supported backend to X11 */
+    gdk_set_allowed_backends ("x11");
+
     tilda_window tw;
     /* NULL set the tw pointers so we can get a clean exit on initialization failure */
     memset(&tw, 0, sizeof(tilda_window));
@@ -683,7 +642,7 @@ int main (int argc, char *argv[])
     lock.pid = getpid ();
     lock.instance = get_instance_number ();
     lock_file = create_lock_file (&lock);
-    config_file = get_config_file_name (lock.instance);
+
     /* End of atomic section */
 
     flock(global_lock.file_descriptor, LOCK_UN);
@@ -703,25 +662,37 @@ int main (int argc, char *argv[])
     textdomain (PACKAGE);
 #endif
 
-
-#ifdef DEBUG
-    /* Have to do this early. */
-    if (getenv ("VTE_PROFILE_MEMORY"))
-        if (atol (getenv ("VTE_PROFILE_MEMORY")) != 0)
-            g_mem_set_vtable (glib_mem_profiler_table);
-#endif
-    /* Start up the configuration system */
-    gint config_init_result = config_init (config_file);
+    config_file = NULL;
 
     /* Parse the command line */
-    need_wizard = parse_cli (argc, argv);
+    tilda_cli_options *cli_options = tilda_cli_options_new ();
+    need_wizard = tilda_cli_options_parse_options (cli_options, argc, argv, &config_file);
+
+    if (config_file) {	  // if there was a config file specified via cli
+        if (!g_file_test (config_file, G_FILE_TEST_EXISTS)) {
+            g_printerr (_("Specified config file '%s' does not exist. Reverting to default path.\n"),
+                    config_file);
+            config_file = get_config_file_name (lock.instance);
+        }
+    } else {    // if not, we look for the defaut config file
+        config_file = get_config_file_name (lock.instance);
+    }
+
+    /* Initialize GTK. Any code that interacts with GTK (e.g. creating a widget)
+     * should come after this call. Gtk initialization should happen before we
+     * initialize the config file. */
+    gtk_init (&argc, &argv);
+
+    /* Start up the configuration system and load from file */
+    gint config_init_result = config_init (config_file);
+
+    /* Set up possible overridden config options */
+    setup_config_from_cli_options(cli_options);
+    g_free(cli_options);
 
     /* We're about to startup X, so set the error handler. */
     XSetErrorHandler (xerror_handler);
 
-    /* Initialize GTK. Any code that interacts with GTK (e.g. creating a widget) should come after this call. */
-    gtk_init (&argc, &argv);
-
     /* This section shows a modal dialog to notify the user that something has gone wrong when loading the config.
      * Earlier version only used to print a message to stderr, but since tilda is usually not started from a
      * console this message would have been lost and a message dialog is much more user friedly.
@@ -741,6 +712,7 @@ int main (int argc, char *argv[])
         gtk_widget_destroy(dialog);
     }
 
+    load_application_css ();
     load_custom_css_file ();
 
     /* create new tilda_window */
@@ -783,7 +755,7 @@ int main (int argc, char *argv[])
      *
      * Note that the key will be bound upon exiting the wizard */
     if (need_wizard) {
-        g_print ("Starting the wizard to configure tilda options.");
+        g_print ("Starting the wizard to configure tilda options.\n");
         wizard (&tw);
     } else {
         gint ret = tilda_keygrabber_bind (config_getstr ("key"), &tw);
@@ -791,7 +763,10 @@ int main (int argc, char *argv[])
         if (!ret)
         {
             /* The key was unbindable, so we need to show the wizard */
-            show_invalid_keybinding_dialog (NULL, _("The keybinding you chose for \"Pull Down Terminal\" is invalid. Please choose another."));
+            const char *message = _("The keybinding you chose for \"Pull Down Terminal\" is invalid. Please choose another.");
+
+            tilda_keybinding_show_invalid_keybinding_dialog (NULL,
+                                                             message);
             wizard (&tw);
         }
     }
@@ -814,4 +789,3 @@ initialization_failed:
     g_free (config_file);
     return 0;
 }
-
