@@ -16,16 +16,11 @@
 #ifndef TILDA_H
 #define TILDA_H
 
-#include "tilda_window.h"
-
 #include <gtk/gtk.h>
 
 G_BEGIN_DECLS
 
 #define TILDA_VERSION PACKAGE_NAME " " PACKAGE_VERSION
-#define QUICK_STRCMP(a, b) (*(a)!=*(b)? \
-    (int) ((unsigned char) *(a) - (unsigned char) *(b)) : \
-    strcmp ((a), (b)))
 
 struct lock_info
 {
@@ -34,8 +29,8 @@ struct lock_info
     gint file_descriptor;
 };
 
-#define GUINT16_TO_FLOAT(color) (color / (double) 0xFFFF)
-#define GUINT16_FROM_FLOAT(value) ((int) (value * 0xFFFF + 0.5))
+#define GUINT16_TO_FLOAT(color) ((color) / (double) 0xFFFF)
+#define GUINT16_FROM_FLOAT(value) ((int) ((value) * 0xFFFF + 0.5))
 
 #define RGB(r,g,b) (r) / (gdouble)G_MAXUINT16, \
                      (g) / (gdouble)G_MAXUINT16, \
@@ -44,13 +39,3 @@ struct lock_info
 G_END_DECLS
 
 #endif
-
-/* Future compatibility with VTE-2.90 */
-#ifndef VTE_290 
-#define vte_terminal_set_colors_rgba vte_terminal_set_colors
-#define vte_terminal_set_color_background_rgba vte_terminal_set_color_background
-#define vte_terminal_set_color_foreground_rgba vte_terminal_set_color_foreground
-#define vte_terminal_set_color_cursor_rgba vte_terminal_set_color_cursor
-#define VteTerminalCursorShape VteCursorShape
-#endif
-
