@@ -19,7 +19,6 @@
 #include "tilda.h"
 #include "tilda_window.h"
 #include "tilda_terminal.h"
-#include "callback_func.h"
 #include "configsys.h"
 #include "wizard.h" /* wizard */
 
@@ -32,20 +31,14 @@
 #include <vte/vte.h>
 #include <string.h>
 
-#define HTTP_REGEXP "(ftp|http)s?://[-a-zA-Z0-9.?$%&/=_~#.,:;+]*"
+#define HTTP_REGEXP "(ftp|http)s?://[\\[\\]-a-zA-Z0-9.?!$%&/=_~#.,:;+]*"
 
-GdkRGBA current_palette[TERMINAL_PALETTE_SIZE];
+static void start_shell (tilda_term *tt, gboolean ignore_custom_command);
+static void start_default_shell (tilda_term *tt);
 
-static gint start_shell (struct tilda_term_ *tt, gboolean ignore_custom_command, const char* working_dir);
 static gint tilda_term_config_defaults (tilda_term *tt);
 
-#ifdef VTE_290
-    static void child_exited_cb (GtkWidget *widget, gpointer data);
-#else
-    static void eof_cb (GtkWidget *widget, gpointer data);
-    /* VTE 2.91 introduced a new status argument to the "child-exited" signal */
-    static void child_exited_cb (GtkWidget *widget, gint status, gpointer data);
-#endif
+static void child_exited_cb (GtkWidget *widget, gint status, gpointer data);
 static void window_title_changed_cb (GtkWidget *widget, gpointer data);
 static int button_press_cb (GtkWidget *widget, GdkEventButton *event, gpointer data);
 static gboolean key_press_cb (GtkWidget *widget, GdkEvent  *event, tilda_term *terminal);
@@ -57,15 +50,27 @@ static void maximize_window_cb (GtkWidget *widget, gpointer data);
 static void restore_window_cb (GtkWidget *widget, gpointer data);
 static void refresh_window_cb (GtkWidget *widget, gpointer data);
 static void move_window_cb (GtkWidget *widget, guint x, guint y, gpointer data);
-#ifdef VTE_290
-static void status_line_changed_cb (GtkWidget *widget, gpointer data);
-#endif
 
-gint tilda_term_free (struct tilda_term_ *term)
+static gchar *get_default_command (void);
+gchar *get_working_directory (tilda_term *terminal);
+
+gint tilda_term_free (tilda_term *term)
 {
     DEBUG_FUNCTION ("tilda_term_free");
     DEBUG_ASSERT (term != NULL);
 
+    g_free (term->initial_working_dir);
+
+    g_signal_handlers_disconnect_by_func (term->vte_term, child_exited_cb, term);
+
+    g_clear_object (&term->hbox);
+    g_clear_object (&term->scrollbar);
+    g_clear_object (&term->vte_term);
+
+    g_regex_unref (term->http_regexp);
+
+    term->http_regexp = NULL;
+
     g_free (term);
 
     return 0;
@@ -85,8 +90,14 @@ static void tilda_terminal_switch_page_cb (GtkNotebook *notebook,
         }
         counter++;
     }
-    const char* current_title = vte_terminal_get_window_title (VTE_TERMINAL (term->vte_term));
-    gtk_window_set_title (GTK_WINDOW (tw->window), current_title);
+
+    char * current_title = tilda_terminal_get_title (term);
+
+    if (current_title != NULL) {
+        gtk_window_set_title (GTK_WINDOW (tw->window), current_title);
+    }
+
+    g_free (current_title);
 }
 
 struct tilda_term_ *tilda_term_init (struct tilda_window_ *tw)
@@ -99,9 +110,8 @@ struct tilda_term_ *tilda_term_init (struct tilda_window_ *tw)
     GError *error = NULL;
     tilda_term *current_tt;
     gint current_tt_index;
-    char *current_tt_dir = NULL;
 
-    term = g_malloc (sizeof (struct tilda_term_));
+    term = g_new0 (tilda_term, 1);
 
     /* Add to GList list of tilda_term structures in tilda_window structure */
     tw->terms = g_list_append (tw->terms, term);
@@ -118,13 +128,16 @@ struct tilda_term_ *tilda_term_init (struct tilda_window_ *tw)
 
     /* Create a non-homogenous hbox, with 0px spacing between members */
     term->hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
+    g_object_ref (term->hbox);
 
     /* Create the terminal */
     term->vte_term = vte_terminal_new ();
+    g_object_ref (term->vte_term);
 
     /* Create the scrollbar for the terminal */
     term->scrollbar = gtk_scrollbar_new (GTK_ORIENTATION_VERTICAL,
         gtk_scrollable_get_vadjustment (GTK_SCROLLABLE (VTE_TERMINAL(term->vte_term))));
+    g_object_ref (term->scrollbar);
 
     gtk_widget_set_no_show_all (term->scrollbar, TRUE);
 
@@ -150,15 +163,6 @@ struct tilda_term_ *tilda_term_init (struct tilda_window_ *tw)
                       G_CALLBACK(child_exited_cb), term);
     g_signal_connect (G_OBJECT(term->vte_term), "window-title-changed",
                       G_CALLBACK(window_title_changed_cb), term);
-#ifdef VTE_290
-    g_signal_connect (G_OBJECT(term->vte_term), "eof",
-                      G_CALLBACK(child_exited_cb), term);
-    g_signal_connect (G_OBJECT(term->vte_term), "status-line-changed",
-                      G_CALLBACK(status_line_changed_cb), term);
-#else
-    g_signal_connect (G_OBJECT(term->vte_term), "eof",
-                      G_CALLBACK(eof_cb), term);
-#endif
     g_signal_connect (G_OBJECT(term->vte_term), "button-press-event",
                       G_CALLBACK(button_press_cb), term);
     g_signal_connect (G_OBJECT(term->vte_term), "key-press-event",
@@ -198,22 +202,13 @@ struct tilda_term_ *tilda_term_init (struct tilda_window_ *tw)
     current_tt = g_list_nth_data (tw->terms, current_tt_index);
     if (current_tt != NULL)
     {
-        current_tt_dir = tilda_term_get_cwd(current_tt);
+        term->initial_working_dir = tilda_term_get_cwd (current_tt);
     }
 
     /* Fork the appropriate command into the terminal */
-    ret = start_shell (term, FALSE, current_tt_dir);
-
-    g_free(current_tt_dir);
-
-    if (ret)
-        goto err_fork;
+    start_shell (term, FALSE);
 
     return term;
-
-err_fork:
-    g_free (term);
-    return NULL;
 }
 
 void tilda_term_set_scrollbar_position (tilda_term *tt, enum tilda_term_scrollbar_positions pos)
@@ -239,50 +234,39 @@ static void window_title_changed_cb (GtkWidget *widget, gpointer data)
     DEBUG_ASSERT (data != NULL);
 
     tilda_term *tt = TILDA_TERM(data);
-    gchar *title = get_window_title (widget);
+    gchar * title = tilda_terminal_get_title (tt);
+    gchar * full_title = tilda_terminal_get_full_title (tt);
     GtkWidget *label;
 
     label = gtk_notebook_get_tab_label (GTK_NOTEBOOK (tt->tw->notebook), tt->hbox);
+
     /* We need to check if the widget that received the title change is the currently
      * active tab. If not we should not update the window title. */
     gint page = gtk_notebook_get_current_page (GTK_NOTEBOOK (tt->tw->notebook));
-    tilda_term *currente_term;
-    gboolean active = FALSE;
-    if (page > 0) {
+
+    gboolean active;
+    if (page >= 0) {
+        tilda_term *currente_term;
         currente_term = g_list_nth_data (tt->tw->terms, (guint) page);
         active = widget == currente_term->vte_term;
-    }
-
-    guint length = (guint) config_getint ("title_max_length");
-
-    if(config_getbool("title_max_length_flag") && strlen(title) > length) {
-        gchar *titleOffset = title + strlen(title) - length;
-        gchar *shortTitle = g_strdup_printf ("...%s", titleOffset);
-        gtk_label_set_text (GTK_LABEL(label), shortTitle);
-        if (active) {
-            gtk_window_set_title (GTK_WINDOW (tt->tw->window), shortTitle);
-        }
-        g_free(shortTitle);
     } else {
-        gtk_label_set_text (GTK_LABEL(label), title);
-        if (active) {
-            gtk_window_set_title (GTK_WINDOW (tt->tw->window), title);
-        }
+        active = TRUE;
     }
 
+    gtk_label_set_text (GTK_LABEL(label), title);
+
+    if (active) {
+        gtk_window_set_title (GTK_WINDOW (tt->tw->window), title);
+    }
+
+    if(config_getbool("show_title_tooltip"))
+      gtk_widget_set_tooltip_text(label, full_title);
+    else
+      gtk_widget_set_tooltip_text(label, "");
+
     g_free (title);
 }
 
-#ifdef VTE_290
-static void status_line_changed_cb (GtkWidget *widget, G_GNUC_UNUSED gpointer data)
-{
-    DEBUG_FUNCTION ("status_line_changed_cb");
-    DEBUG_ASSERT (widget != NULL);
-
-    g_print ("Status = `%s'.\n", vte_terminal_get_status_line (VTE_TERMINAL(widget)));
-}
-#endif
-
 static void iconify_window_cb (G_GNUC_UNUSED GtkWidget *widget, gpointer data)
 {
     DEBUG_FUNCTION ("iconify_window_cb");
@@ -418,14 +402,43 @@ char* tilda_term_get_cwd(struct tilda_term_* tt)
     return cwd;
 }
 
+static void
+shell_spawned_cb (VteTerminal *terminal,
+                  GPid pid,
+                  GError *error,
+                  gpointer user_data)
+{
+    DEBUG_FUNCTION ("shell_spawned_cb");
+
+    tilda_term *tt;
+
+    tt = user_data;
+
+    if (error)
+    {
+        if (tt->dropped_to_default_shell)
+        {
+            g_printerr (_("Unable to launch default shell: %s\n"), get_default_command ());
+        } else {
+            g_printerr (_("Unable to launch custom command: %s\n"), config_getstr ("command"));
+            g_printerr (_("Launching custom command failed with error: %s\n"), error->message);
+            g_printerr (_("Launching default shell instead\n"));
+
+            start_default_shell (tt);
+            tt->dropped_to_default_shell = TRUE;
+        }
+
+        return;
+    }
+
+    tt->pid = pid;
+}
+
 /* Fork a shell into the VTE Terminal
  *
  * @param tt the tilda_term to fork into
- *
- * SUCCESS: return 0
- * FAILURE: return non-zero
  */
-static gint start_shell (struct tilda_term_ *tt, gboolean ignore_custom_command, const char* working_dir)
+static void start_shell (tilda_term *tt, gboolean ignore_custom_command)
 {
     DEBUG_FUNCTION ("start_shell");
     DEBUG_ASSERT (tt != NULL);
@@ -435,13 +448,6 @@ static gint start_shell (struct tilda_term_ *tt, gboolean ignore_custom_command,
     gchar **argv;
     GError *error = NULL;
 
-    gchar *default_command;
-
-    if (working_dir == NULL || config_getbool ("inherit_working_dir") == FALSE)
-    {
-        working_dir = config_getstr ("working_dir");
-    }
-
     if (config_getbool ("run_command") && !ignore_custom_command)
     {
         ret = g_shell_parse_argv (config_getstr ("command"), &argc, &argv, &error);
@@ -453,54 +459,43 @@ static gint start_shell (struct tilda_term_ *tt, gboolean ignore_custom_command,
             g_printerr (_("Launching default shell instead\n"));
 
             g_error_free (error);
-            goto launch_default_shell;
+
+            start_default_shell (tt);
         }
 
+        gchar *working_dir = get_working_directory (tt);
+        gint command_timeout = config_getint ("command_timeout_ms");
+
         char **envv = malloc(2*sizeof(void *));
         envv[0] = getenv("PATH");
         envv[1] = NULL;
-#ifdef VTE_290
-        ret = vte_terminal_fork_command_full (VTE_TERMINAL (tt->vte_term),
-            VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
-            working_dir, /* const char *working_directory */
-            argv, /* char **argv */
-            envv, /* char **envv */
-            G_SPAWN_SEARCH_PATH,    /* GSpawnFlags spawn_flags */
-            NULL, /* GSpawnChildSetupFunc child_setup */
-            NULL, /* gpointer child_setup_data */
-            &tt->pid, /* GPid *child_pid */
-            NULL  /* GError **error */
-            );
-#else
-        ret = vte_terminal_spawn_sync (VTE_TERMINAL (tt->vte_term),
-            VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
-            working_dir, /* const char *working_directory */
-            argv, /* char **argv */
-            envv, /* char **envv */
-            G_SPAWN_SEARCH_PATH,    /* GSpawnFlags spawn_flags */
-            NULL, /* GSpawnChildSetupFunc child_setup */
-            NULL, /* gpointer child_setup_data */
-            &tt->pid, /* GPid *child_pid */
-            NULL, /* GCancellable * cancellable, */
-            NULL  /* GError **error */
-            );
-#endif
+
+        vte_terminal_spawn_async (VTE_TERMINAL (tt->vte_term),
+                                  VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
+                                  working_dir, /* const char *working_directory */
+                                  argv, /* char **argv */
+                                  envv, /* char **envv */
+                                  G_SPAWN_SEARCH_PATH,    /* GSpawnFlags spawn_flags */
+                                  NULL, /* GSpawnChildSetupFunc child_setup */
+                                  NULL, /* gpointer child_setup_data */
+                                  NULL, /* GDestroyNotify child_setup_data_destroy */
+                                  command_timeout, /* timeout in ms */
+                                  NULL, /* GCancellable * cancellable, */
+                                  shell_spawned_cb,  /* VteTerminalSpawnAsyncCallback callback */
+                                  tt);   /* user_data */
+
         g_strfreev (argv);
         g_free (envv);
-
-        /* Check for error */
-        if (ret == FALSE)
-        {
-            g_printerr (_("Unable to launch custom command: %s\n"), config_getstr ("command"));
-            g_printerr (_("Launching default shell instead\n"));
-
-            goto launch_default_shell;
-        }
-
-        return 0; /* SUCCESS: the early way out */
+    } else {
+        start_default_shell (tt);
     }
+}
+
+static void
+start_default_shell (tilda_term *tt)
+{
+    gchar  **argv;
 
-launch_default_shell:
 
     /* If we have dropped to the default shell before, then this time, we
      * do not spawn a new shell, but instead close the current shell. This will
@@ -510,18 +505,13 @@ launch_default_shell:
         gint index = gtk_notebook_page_num (GTK_NOTEBOOK(tt->tw->notebook),
             tt->hbox);
         tilda_window_close_tab (tt->tw, index, FALSE);
-        return 0;
-    }
-    if (ignore_custom_command) {
-        tt->dropped_to_default_shell = TRUE;
+
+        return;
     }
 
-    /* No custom command, get it from the environment */
-    default_command = (gchar *) g_getenv ("SHELL");
-
-    /* Check for error */
-    if (default_command == NULL)
-        default_command = "/bin/sh";
+    gchar *default_command = get_default_command ();
+    gchar *working_dir = get_working_directory (tt);
+    gint command_timeout = config_getint ("command_timeout_ms");
 
     /* We need to create a NULL terminated list of arguments.
      * The first item is the command to execute in the shell, in this
@@ -544,65 +534,60 @@ launch_default_shell:
         argv[0] = default_command;
         argv[1] = NULL;
     }
-#ifdef VTE_290
-    ret = vte_terminal_fork_command_full (VTE_TERMINAL (tt->vte_term),
-        VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
-        working_dir, /* const char *working_directory */
-        argv, /* char **argv */
-        NULL, /* char **envv */
-        flags,    /* GSpawnFlags spawn_flags */
-        NULL, /* GSpawnChildSetupFunc child_setup */
-        NULL, /* gpointer child_setup_data */
-        &tt->pid, /* GPid *child_pid */
-        NULL  /* GError **error */
-        );
-#else
-    ret = vte_terminal_spawn_sync (VTE_TERMINAL (tt->vte_term),
-        VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
-        working_dir, /* const char *working_directory */
-        argv, /* char **argv */
-        NULL, /* char **envv */
-        flags,    /* GSpawnFlags spawn_flags */
-        NULL, /* GSpawnChildSetupFunc child_setup */
-        NULL, /* gpointer child_setup_data */
-        &tt->pid, /* GPid *child_pid */
-        NULL, /* GCancellable *cancellable */
-        NULL  /* GError **error */
-        );
-#endif
+
+    vte_terminal_spawn_async (VTE_TERMINAL (tt->vte_term),
+                              VTE_PTY_DEFAULT, /* VtePtyFlags pty_flags */
+                              working_dir, /* const char *working_directory */
+                              argv, /* char **argv */
+                              NULL, /* char **envv */
+                              flags,    /* GSpawnFlags spawn_flags */
+                              NULL, /* GSpawnChildSetupFunc child_setup */
+                              NULL, /* gpointer child_setup_data */
+                              NULL, /* GDestroyNotify child_setup_data_destroy */
+                              command_timeout, /* timeout in ms */
+                              NULL, /* GCancellable * cancellable, */
+                              shell_spawned_cb,  /* VteTerminalSpawnAsyncCallback callback */
+                              tt);   /* user_data */
+
     g_free(argv1);
     g_free (argv);
+}
 
-    if (ret == -1)
+gchar *
+get_default_command ()
+{
+    /* No custom command, get it from the environment */
+    gchar *default_command = (gchar *) g_getenv ("SHELL");
+
+    /* Check for error */
+    if (default_command == NULL)
+        default_command = "/bin/sh";
+
+    return default_command;
+}
+
+gchar *get_working_directory (tilda_term *terminal)
+{
+    gchar *working_dir;
+
+    working_dir = terminal->initial_working_dir;
+
+    if (working_dir == NULL || config_getbool ("inherit_working_dir") == FALSE)
     {
-        g_printerr (_("Unable to launch default shell: %s\n"), default_command);
-        return ret;
+        working_dir = config_getstr ("working_dir");
     }
 
-    return 0;
+    return working_dir;
 }
 
-#ifdef VTE_291
-static void eof_cb (GtkWidget *widget, gpointer data) {
-    DEBUG_FUNCTION ("eof_cb");
-    DEBUG_ASSERT (widget != NULL);
-    DEBUG_ASSERT (data != NULL);
-
-    child_exited_cb (widget, 0, data);
-}
-#endif
-
-#ifdef VTE_290
-static void child_exited_cb (GtkWidget *widget, gpointer data)
-#else
 static void child_exited_cb (GtkWidget *widget, gint status, gpointer data)
-#endif
 {
     DEBUG_FUNCTION ("child_exited_cb");
     DEBUG_ASSERT (widget != NULL);
     DEBUG_ASSERT (data != NULL);
 
     tilda_term *tt = TILDA_TERM(data);
+
     gint index = gtk_notebook_page_num (GTK_NOTEBOOK(tt->tw->notebook), tt->hbox);
 
     /* Make sure we got a valid index */
@@ -625,10 +610,13 @@ static void child_exited_cb (GtkWidget *widget, gint status, gpointer data)
             break;
         case RESTART_COMMAND:
             vte_terminal_feed (VTE_TERMINAL(tt->vte_term), "\r\n\r\n", 4);
-            start_shell (tt, FALSE, NULL);
+            start_shell (tt, FALSE);
             break;
         case DROP_TO_DEFAULT_SHELL:
-            start_shell (tt, TRUE, NULL);
+            tt->initial_working_dir = NULL;
+            start_default_shell (tt);
+            tt->dropped_to_default_shell = TRUE;
+            break;
         default:
             break;
     }
@@ -647,20 +635,16 @@ static gint tilda_term_config_defaults (tilda_term *tt)
     DEBUG_FUNCTION ("tilda_term_config_defaults");
     DEBUG_ASSERT (tt != NULL);
     GdkRGBA fg, bg, cc;
+    GdkRGBA *current_palette;
     gchar* word_chars;
-    gint i;
     gint cursor_shape;
 
     /** Colors & Palette **/
     bg.red   =    GUINT16_TO_FLOAT(config_getint ("back_red"));
     bg.green =    GUINT16_TO_FLOAT(config_getint ("back_green"));
     bg.blue  =    GUINT16_TO_FLOAT(config_getint ("back_blue"));
-#ifdef VTE_290
-    bg.alpha =    1.0;
-    gdouble transparency_level = 0.0;
-#else
+
     bg.alpha =    (config_getbool("enable_transparency") ? GUINT16_TO_FLOAT(config_getint ("back_alpha")) : 1.0);
-#endif
 
     fg.red   =    GUINT16_TO_FLOAT(config_getint ("text_red"));
     fg.green =    GUINT16_TO_FLOAT(config_getint ("text_green"));
@@ -672,36 +656,39 @@ static gint tilda_term_config_defaults (tilda_term *tt)
     cc.blue  =    GUINT16_TO_FLOAT(config_getint ("cursor_blue"));
     cc.alpha = 1.0;
 
-    for(i = 0;i < TERMINAL_PALETTE_SIZE; i++) {
+    current_palette = tilda_palettes_get_current_palette ();
+
+    for(guint i = 0; i < TILDA_COLOR_PALETTE_SIZE; i++) {
         current_palette[i].red   = GUINT16_TO_FLOAT(config_getnint ("palette", i*3));
         current_palette[i].green = GUINT16_TO_FLOAT(config_getnint ("palette", i*3+1));
         current_palette[i].blue  = GUINT16_TO_FLOAT(config_getnint ("palette", i*3+2));
         current_palette[i].alpha = 1.0;
     }
 
-    vte_terminal_set_colors_rgba (VTE_TERMINAL(tt->vte_term), &fg, &bg, current_palette, TERMINAL_PALETTE_SIZE);
+    vte_terminal_set_colors (VTE_TERMINAL(tt->vte_term),
+                             &fg,
+                             &bg,
+                             current_palette,
+                             TILDA_COLOR_PALETTE_SIZE);
 
     /** Bells **/
     vte_terminal_set_audible_bell (VTE_TERMINAL(tt->vte_term), config_getbool ("bell"));
-#ifdef VTE_290
-    vte_terminal_set_visible_bell (VTE_TERMINAL(tt->vte_term), config_getbool ("bell"));
-#endif
+
     /** Cursor **/
     vte_terminal_set_cursor_blink_mode (VTE_TERMINAL(tt->vte_term),
             (config_getbool ("blinks"))?VTE_CURSOR_BLINK_ON:VTE_CURSOR_BLINK_OFF);
-    vte_terminal_set_color_cursor_rgba (VTE_TERMINAL(tt->vte_term), &cc);
+    vte_terminal_set_color_cursor (VTE_TERMINAL(tt->vte_term), &cc);
+    vte_terminal_set_color_cursor_foreground (VTE_TERMINAL(tt->vte_term), &bg);
 
     cursor_shape = config_getint("cursor_shape");
     if (cursor_shape < 0 || cursor_shape > 2) {
         config_setint("cursor_shape", 0);
         cursor_shape = 0;
     }
-    vte_terminal_set_cursor_shape(VTE_TERMINAL(tt->vte_term), (VteTerminalCursorShape)cursor_shape);
+    vte_terminal_set_cursor_shape(VTE_TERMINAL(tt->vte_term),
+                                  (VteCursorShape) cursor_shape);
 
     /** Scrolling **/
-#ifdef VTE_290
-    vte_terminal_set_scroll_background (VTE_TERMINAL(tt->vte_term), config_getbool ("scroll_background"));
-#endif
     vte_terminal_set_scroll_on_output (VTE_TERMINAL(tt->vte_term), config_getbool ("scroll_on_output"));
     vte_terminal_set_scroll_on_keystroke (VTE_TERMINAL(tt->vte_term), config_getbool ("scroll_on_key"));
 
@@ -709,8 +696,6 @@ static gint tilda_term_config_defaults (tilda_term *tt)
     vte_terminal_set_mouse_autohide (VTE_TERMINAL(tt->vte_term), FALSE); /* TODO: make this configurable */
 
     /** Text Properties **/
-    vte_terminal_set_allow_bold (VTE_TERMINAL(tt->vte_term), config_getbool ("bold"));
-    gtk_widget_set_double_buffered (tt->vte_term, config_getbool("double_buffer"));
     PangoFontDescription *description =
         pango_font_description_from_string (config_getstr ("font"));
     vte_terminal_set_font (VTE_TERMINAL (tt->vte_term), description);
@@ -759,33 +744,8 @@ static gint tilda_term_config_defaults (tilda_term *tt)
         word_chars = DEFAULT_WORD_CHARS;
     }
 
-    /**
-     * The word_chars feature was removed from VTE in 38.x and reintroduced with a different API function in VTE
-     * 40.x, so we need the following compile guards to ensure we only include the word_chars feature in the
-     * supported versions.
-     **/
+    vte_terminal_set_word_char_exceptions (VTE_TERMINAL (tt->vte_term), word_chars);
 
-#ifdef VTE_290 /* VTE 2.90 only */
-    vte_terminal_set_word_chars (VTE_TERMINAL(tt->vte_term), word_chars);
-
-    /** Background **/
-    if (config_getbool ("use_image"))
-        vte_terminal_set_background_image_file (VTE_TERMINAL(tt->vte_term), config_getstr ("image"));
-    else
-        vte_terminal_set_background_image_file (VTE_TERMINAL(tt->vte_term), NULL);
-    transparency_level = ((gdouble) config_getint ("transparency"))/100;
-
-    if (config_getbool ("enable_transparency") && transparency_level > 0)
-    {
-        vte_terminal_set_background_saturation (VTE_TERMINAL (tt->vte_term), transparency_level);
-        vte_terminal_set_opacity (VTE_TERMINAL (tt->vte_term), (1.0 - transparency_level) * 0xffff);
-        vte_terminal_set_background_transparent (VTE_TERMINAL(tt->vte_term), !tt->tw->have_argb_visual);
-    }
-#else
-    #if VTE_MINOR_VERSION >= 40
-        vte_terminal_set_word_char_exceptions (VTE_TERMINAL (tt->vte_term), word_chars);
-    #endif
-#endif
     return 0;
 }
 
@@ -799,7 +759,7 @@ menu_copy_cb (GSimpleAction *action,
 
     tilda_term *tt = TILDA_TERM(user_data);
 
-    vte_terminal_copy_clipboard (VTE_TERMINAL (tt->vte_term));
+    vte_terminal_copy_clipboard_format (VTE_TERMINAL (tt->vte_term), VTE_FORMAT_TEXT);
 }
 
 static void
@@ -835,7 +795,7 @@ menu_quit_cb (GSimpleAction *action,
 {
     DEBUG_FUNCTION ("menu_quit_cb");
 
-    gtk_main_quit ();
+    tilda_window_confirm_quit(TILDA_WINDOW(user_data));
 }
 
 static void
@@ -868,7 +828,7 @@ menu_searchbar_cb(GSimpleAction *action,
     DEBUG_FUNCTION ("menu_fullscreen_cb");
     DEBUG_ASSERT (user_data != NULL);
 
-    toggle_searchbar_cb (TILDA_WINDOW(user_data));
+    tilda_window_toggle_searchbar (TILDA_WINDOW(user_data));
 }
 
 static void
@@ -882,72 +842,25 @@ menu_close_tab_cb (GSimpleAction *action,
     tilda_window_close_current_tab (TILDA_WINDOW(user_data));
 }
 
-static void on_popup_hide (GtkWidget *widget, gpointer data)
+static void on_popup_hide (GtkWidget *widget, tilda_window *tw)
 {
     DEBUG_FUNCTION("on_popup_hide");
     DEBUG_ASSERT(widget != NULL);
-    DEBUG_ASSERT(data != NULL);
+    DEBUG_ASSERT(tw != NULL);
 
-    tilda_window *tw = TILDA_WINDOW(data);
     tw->disable_auto_hide = FALSE;
 }
 
-static void popup_menu (tilda_window *tw, tilda_term *tt)
+static void
+popup_menu (tilda_window *tw, tilda_term *tt, GdkEvent * event)
 {
     DEBUG_FUNCTION ("popup_menu");
     DEBUG_ASSERT (tw != NULL);
     DEBUG_ASSERT (tt != NULL);
 
-    static const gchar menu_str[] = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
-            "<interface>\n"
-            "  <menu id=\"menu\">\n"
-            "    <section>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_New Tab</attribute>\n"
-            "        <attribute name=\"action\">window.new-tab</attribute>\n"
-            "      </item>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Close Tab</attribute>\n"
-            "        <attribute name=\"action\">window.close-tab</attribute>\n"
-            "      </item>\n"
-            "    </section>\n"
-            "    <section>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Copy</attribute>\n"
-            "        <attribute name=\"action\">window.copy</attribute>\n"
-            "      </item>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Paste</attribute>\n"
-            "        <attribute name=\"action\">window.paste</attribute>\n"
-            "      </item>\n"
-            "    </section> \n"
-            "    <section>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Toggle Fullscreen</attribute>\n"
-            "        <attribute name=\"action\">window.fullscreen</attribute>\n"
-            "      </item>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Toggle Searchbar</attribute>\n"
-            "        <attribute name=\"action\">window.searchbar</attribute>\n"
-            "      </item>\n"
-            "    </section> \n"
-            "    <section>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Preferences</attribute>\n"
-            "        <attribute name=\"action\">window.preferences</attribute>\n"
-            "      </item>\n"
-            "    </section> \n"
-            "    <section>\n"
-            "      <item>\n"
-            "        <attribute name=\"label\" translatable=\"yes\">_Quit</attribute>\n"
-            "        <attribute name=\"action\">window.quit</attribute>\n"
-            "      </item>\n"
-            "    </section> \n"
-            "  </menu>\n"
-            "</interface>";
-
     GtkBuilder *builder = gtk_builder_new();
-    gtk_builder_add_from_string(builder, menu_str, strlen(menu_str), NULL);
+
+    gtk_builder_add_from_resource (builder, "/org/tilda/menu.ui", NULL);
 
     /* Create the action group */
     GSimpleActionGroup *action_group = g_simple_action_group_new();
@@ -993,9 +906,7 @@ static void popup_menu (tilda_window *tw, tilda_term *tt)
     tw->disable_auto_hide = TRUE;
     g_signal_connect (G_OBJECT(menu), "unmap", G_CALLBACK(on_popup_hide), tw);
 
-    /* Display the menu */
-    gtk_menu_popup (GTK_MENU(menu), NULL, NULL, NULL, NULL, 3, gtk_get_current_event_time());
-    gtk_widget_show_all(menu);
+    gtk_menu_popup_at_pointer (GTK_MENU (menu), event);
 }
 
 static int button_press_cb (G_GNUC_UNUSED GtkWidget *widget, GdkEventButton *event, gpointer data)
@@ -1007,7 +918,6 @@ static int button_press_cb (G_GNUC_UNUSED GtkWidget *widget, GdkEventButton *eve
     tilda_term *tt;
     gchar *match;
     gint tag;
-    gint ypad;
     gchar *cmd;
     gchar *web_browser_cmd;
     gboolean ret = FALSE;
@@ -1023,38 +933,47 @@ static int button_press_cb (G_GNUC_UNUSED GtkWidget *widget, GdkEventButton *eve
             tilda_window_prev_tab (tt->tw);
             break;
         case 3: /* Right Click */
-            popup_menu (tt->tw, tt);
+            popup_menu (tt->tw, tt, (GdkEvent *) event);
             break;
         case 2: /* Middle Click */
             break;
         case 1: /* Left Click */
             terminal  = VTE_TERMINAL(tt->vte_term);
-            ypad = gtk_widget_get_margin_bottom(GTK_WIDGET(terminal));
+
+#if VTE_CHECK_VERSION(0, 38, 0)
+            match = vte_terminal_match_check_event (terminal,
+                                                    (GdkEvent *) event,
+                                                    &tag);
+#else
+            gint ypad = gtk_widget_get_margin_bottom(GTK_WIDGET(terminal));
+
+            glong column = (glong) ((event->x - ypad) /
+                                    vte_terminal_get_char_width (terminal));
+            glong row = (glong) ((event->y - ypad) /
+                                 vte_terminal_get_char_height (terminal));
 
             match = vte_terminal_match_check (terminal,
-                    (event->x - ypad) /
-                    vte_terminal_get_char_width (terminal),
-                    (event->y - ypad) /
-                    vte_terminal_get_char_height (terminal),
-                    &tag);
+                                              column,
+                                              row,
+                                              &tag);
+#endif
 
             /* Check if we can launch a web browser, and do so if possible */
             if (match != NULL)
             {
-#if DEBUG
-                g_print ("Got a Left Click -- Matched: `%s' (%d)\n", match, tag);
-#endif
+                g_debug ("Got a Left Click -- Matched: `%s' (%d)", match, tag);
+
                 web_browser_cmd = g_strescape (config_getstr ("web_browser"), NULL);
                 cmd = g_strdup_printf ("%s %s", web_browser_cmd, match);
-#if DEBUG
-                g_print ("Launching command: `%s'\n", cmd);
-#endif
+
+                g_debug ("Launching command: `%s'", cmd);
+
                 ret = g_spawn_command_line_async(cmd, NULL);
 
                 /* Check that the command launched */
                 if (!ret)
                 {
-                    g_printerr (_("Failed to launch the web browser. The command was `%s'\n"), cmd);
+                    g_critical (_("Failed to launch the web browser. The command was `%s'\n"), cmd);
                     TILDA_PERROR ();
                 }
 
@@ -1078,10 +997,94 @@ gboolean key_press_cb (GtkWidget *widget,
     if(event->type == GDK_KEY_PRESS) {
         GdkEventKey *keyevent = (GdkEventKey*) event;
         if(keyevent->keyval == GDK_KEY_Menu) {
-            popup_menu(terminal->tw, terminal);
+            popup_menu(terminal->tw, terminal, event);
         }
     }
     return GDK_EVENT_PROPAGATE;
 }
 
-/* vim: set ts=4 sts=4 sw=4 expandtab: */
+gchar * tilda_terminal_get_full_title (tilda_term *tt)
+{
+    DEBUG_FUNCTION ("tilda_terminal_get_title");
+
+    const gchar *vte_title;
+    gchar *window_title;
+    gchar *initial;
+    gchar *title;
+
+    vte_title = vte_terminal_get_window_title (VTE_TERMINAL (tt->vte_term));
+    window_title = g_strdup (vte_title);
+    initial = g_strdup (config_getstr ("title"));
+
+    /* These are not needed anywhere else. If they ever are, move them to a header file */
+    enum d_set_title { NOT_DISPLAYED, AFTER_INITIAL, BEFORE_INITIAL, REPLACE_INITIAL };
+
+    switch (config_getint ("d_set_title"))
+    {
+        case REPLACE_INITIAL:
+            title = (window_title != NULL) ? g_strdup (window_title)
+                                           : g_strdup (_("Untitled"));
+            break;
+
+        case BEFORE_INITIAL:
+            title = (window_title != NULL) ? g_strdup_printf ("%s - %s", window_title, initial)
+                                           : g_strdup (initial);
+            break;
+
+        case AFTER_INITIAL:
+            title = (window_title != NULL) ? g_strdup_printf ("%s - %s", initial, window_title)
+                                           : g_strdup (initial);
+            break;
+
+        case NOT_DISPLAYED:
+            title = g_strdup (initial);
+            break;
+
+        default:
+            g_printerr (_("Bad value for \"d_set_title\" in config file\n"));
+            title = g_strdup ("");
+            break;
+    }
+
+    g_free (window_title);
+    g_free (initial);
+
+    return title;
+}
+
+gchar *tilda_terminal_get_title (tilda_term *tt)
+{
+    DEBUG_FUNCTION ("tilda_terminal_get_title");
+    DEBUG_ASSERT (tt != NULL);
+
+    gchar * title;
+    gchar * final_title;
+
+    title = tilda_terminal_get_full_title (tt);
+
+    /* These are not needed anywhere else. If they ever are, move them to a header file */
+    enum { SHOW_FULL_TITLE, SHOW_FIRST_N_CHARS, SHOW_LAST_N_CHARS };
+
+    guint max_length = (guint) config_getint ("title_max_length");
+    guint title_behaviour = config_getint("title_behaviour");
+
+    if (strlen (title) > max_length) {
+        if (title_behaviour == SHOW_FULL_TITLE) {
+            final_title = g_strdup (title);
+        } else if (title_behaviour == SHOW_FIRST_N_CHARS) {
+            final_title = g_strdup_printf ("%.*s...", max_length, title);
+        } else if (title_behaviour == SHOW_LAST_N_CHARS) {
+            gchar *titleOffset = title + strlen(title) - max_length;
+            final_title = g_strdup_printf ("...%s", titleOffset);
+        } else {
+            g_printerr ("Bad value for \"title_behaviour\" in config file.\n");
+            final_title = g_strdup (title);
+        }
+    } else {
+        final_title = g_strdup (title);
+    }
+
+    g_free (title);
+
+    return final_title;
+}
